#include "stdafx.h"
#include <assert.h>
#include <ctype.h>
#include <stdio.h>
#include <stdarg.h>
#include <time.h>

#ifdef _WIN32
#  pragma warning(disable:4786)
#endif

#include <map>
#include <vector>

#define STANDALONE 1

#undef CONST
#undef OUT

//------------------------------------------------------------------------------

const char *timeStamp(bool longForm=true)
{
   const time_t timer = time(NULL);
   static char  formattedTime[26];

   struct tm *tm_ptr = gmtime(&timer);
   int        gmt_dy = tm_ptr->tm_yday;
   int        gmt_hr = tm_ptr->tm_hour;
   int        gmt_mn = tm_ptr->tm_min;

   tm_ptr = localtime(&timer);
   if (gmt_dy > tm_ptr->tm_yday) gmt_hr += 24;

   if (longForm) {
      int n = strftime(formattedTime, sizeof(formattedTime)-1, "%Y-%m-%dT%H:%M:%S", tm_ptr);
      if (gmt_mn > tm_ptr->tm_min) {
         gmt_mn -= 60;
         gmt_hr +=  1;
      }
      sprintf(formattedTime+n, "%+03d:%02d", tm_ptr->tm_hour-gmt_hr, tm_ptr->tm_min-gmt_mn);
   }
   else {
      // Useful when generating time-stamped file names.
      strftime(formattedTime, sizeof(formattedTime)-1, "%Y%m%d%H%M", tm_ptr);
   }

   return formattedTime;
}

//------------------------------------------------------------------------------

class listFile {
   public:
      listFile(CString sourceName)
       : _currentLine(0)
      {
         _baseName = sourceName.Left(sourceName.ReverseFind('.'));
         _lst      = fopen(LPCTSTR((_baseName+".lst")), "w");
      }
     ~listFile()
      {
         fclose(_lst);
      }

      const CString &baseName() const { return _baseName; }

      void setLine(int l) { _thisLine = l; }

      void write(int lnum, int prio, CString format, ...)
      {
         if (_lst) {
            _thisLine = lnum;
            if (_thisLine != _currentLine) flush();
            _currentLine = _thisLine;

            va_list args;
            va_start(args, *reinterpret_cast<char *>(&format));
               CString output;
               output.FormatV(LPCTSTR(format), args);
               _lineBuffer.push_back(line(_thisLine, prio, output));
            va_end(args);
         }
      }

      void write(int prio, CString format, ...)
      {
         if (_lst) {
            if (_thisLine != _currentLine) flush();
            _currentLine = _thisLine;

            va_list args;
            va_start(args, *reinterpret_cast<char *>(&format));
               CString output;
               output.FormatV(LPCTSTR(format), args);
               _lineBuffer.push_back(line(_thisLine, prio, output));
            va_end(args);
         }
      }

      void flush()
      {
         if (_lst) {
            for (int p = 0; p < 10; p++) {
               for (unsigned i = 0; i < _lineBuffer.size(); i++) {
                  if (_lineBuffer[i].prio == p) {
                     fprintf(_lst, "%s", LPCTSTR(_lineBuffer[i].text));
                     //fprintf(_lst, "%d'%d'%s", _lineBuffer[i].lnum,_lineBuffer[i].prio,LPCTSTR(_lineBuffer[i].text));
                  }
               }
            }
            fflush(_lst);
            _lineBuffer.clear();
         }
      }

   private:
      int          _thisLine;
      int          _currentLine;
      FILE        *_lst;
      CString      _baseName;

      struct line {
         line(int l, int p, CString t) : lnum(l), prio(p), text(t) { }
         int     lnum;
         int     prio;
         CString text;
      };

      std::vector<line> _lineBuffer;
};

//------------------------------------------------------------------------------

class exprError {
   const char *file;
   int         lineNumber;
   char        tok[128];
   char        str[128];
   char        title[128];
   bool        found;

   public:
      exprError() : file(NULL), lineNumber(0), found(false) { tok[0] = title[0] = str[0] = '\0'; }

      void setFile (const char *f) { file = f;         }
      void setTitle(const char *t) { strcpy(title, t); }
      void setLine (int         l) { lineNumber = l;   }

      void reset(char    c) { tok[0] = c; tok[1] = '\0'; }
      void reset(CString s) { strcpy(tok, LPCTSTR(s)); }

      const char *errorString() { return str; }
      int         errorLine  () { return lineNumber; }

      char append(char c)
      {
         int len    = strlen(tok);
         tok[len]   = c;
         tok[len+1] = '\0';
         return c;
      }

      int backup()
      {
         int len = strlen(tok);
         if (len) tok[len-1] = '\0';
         return len;
      }

      void Error(const char *Msg, ...)
      {
         va_list args;
         va_start(args, *reinterpret_cast<char *>(&Msg));
            CString output;
            output.FormatV(Msg, args);
         va_end(args);

         found = true;
         if (lineNumber == 0)
            sprintf(str, "ERROR: %s '%s'", LPCTSTR(output), tok);
         else
            sprintf(str, "%s:%d: %s '%s'", file?file:"???", lineNumber, LPCTSTR(output), tok);
         throw *this;
      }

      void display()
      {
#ifdef STANDALONE
         fprintf(stderr, "%s\n", str);
#else
         char outStr[512];
         sprintf(outStr, "%s\n\nTerminate MegaTune?", str);
         if (MessageBoxA(NULL, outStr, title, MB_YESNO | MB_APPLMODAL) == IDYES) {
            exit(1);
         }
#endif
      }
};

//------------------------------------------------------------------------------

class symbol {
   public:
      typedef enum { variable, function } Type;
      typedef enum { fNone, fARR, fCRV, fTBL } fType;
      typedef enum { eNone, eConst, eOut, eLast } Exposure;
      static  char *expNames[];

      symbol(CString &name)
       : _name      (name)
       , _id        (_next++)
       , _type      (variable)
       , _ftype     (fNone)
       , _args      (0)
       , _lhs       (false)
       , _totalBytes(2)
       , _exposed   (eNone)
       , _addr      (0)
      { }


      int         id   () { return _id;    }
      const char *name () { return LPCTSTR(_name); }
      Type        type () { return _type;  }
      fType       ftype() { return _ftype; }
      int         args () { return _args;  }
      int         size () { return _size;  }
      int         bytes() { return _totalBytes; }

      unsigned    addr()           { return _addr; }
      void        addr(unsigned a) { _addr = a;    }

      bool        lhs()       { return _lhs; }
      void        lhs(bool l) { _lhs = l;    }

      Exposure    exposed() { return _exposed; }

      void        type(Type t) { _type = t; }
      void        args(int  a) { _args = a; }

      void ARR(int s)
      {
         _ftype = fARR;
         type(symbol::function);
         args(1);
         _size = s;
         _totalBytes = 2 * _size;
      }
      void CRV(const symbol *x)
      {
         _ftype = fCRV;
         type(symbol::function);
         args(1); // 2?, see ::call in aFactor
         xBins = x;
         if (xBins->_ftype != fARR) throw "Curve X must be an array";
         _totalBytes = xBins->_totalBytes;
      }
      void TBL(const symbol *x, const symbol *y)
      {
         _ftype = fTBL;
         type(symbol::function);
         args(2);
         xBins = x;
         yBins = y;
         if (xBins->_ftype != fARR) throw "Table X must be an array";
         if (yBins->_ftype != fARR) throw "Table Y must be an array";
         _totalBytes = 2 * xBins->_size * yBins->_size;
      }
      void CONST(CString &u, double s, double t, double l, double h, int d)
      {
         _exposed = eConst;
         units  = u;
         scale  = s;
         trans  = t;
         lo     = l;
         hi     = h;
         digits = d;
      }
      void OUT(CString &u, double s, double t, int d)
      {
         _exposed = eOut;
         units  = u;
         scale  = s;
         trans  = t;
         digits = d;
      }

   private:
      CString    _name;
      int        _id;
      Type       _type;
      fType      _ftype;
      int        _args;
      bool       _lhs;
      int        _totalBytes;
      Exposure   _exposed;

      // Funky crap for built-in data structuring "functions"
      int           _size;
      unsigned      _addr;
      const symbol *xBins;
      const symbol *yBins;
      CString       units;
      double        scale;
      double        trans;
      double        hi;
      double        lo;
      int           digits;

      static int   _next;
};

char *symbol::expNames[] = {
   "Private Symbols ----------------------------------------------------------------",
   "[Constants] --------------------------------------------------------------------",
   "[OutputChannels] ---------------------------------------------------------------",
   ""
};

int symbol::_next = 0;

class symbolTable {
   public:
      symbolTable() {
         add("TIME");
         add("ACC",   1, 5);
         add("DERIV", 2, 3);
         add("MAX",   3, 2);
         add("MIN",   4, 2);
         add("PID",   5,99);
         add("WHEN",  6, 4);
      }
     ~symbolTable() {
         for (std::map<CString, symbol *>::iterator i = _table.begin(); i != _table.end(); i++) {
            delete (*i).second;
         }
      }

      symbol *lookup(CString &symName) {
         if (_table.find(symName) != _table.end()) return _table[symName];
         return NULL;
      }

      symbol *add(CString symName) {
         return _table[symName] = new symbol(symName);
      }

      void computeAddresses()
      {
         std::vector<symbol *> s(_table.size());
         for (std::map<CString, symbol *>::iterator i = _table.begin(); i != _table.end(); i++) {
            s[(*i).second->id()] = (*i).second;
         }
         int ramAddr = 0x3000;
         int flhAddr = 0x4020;
         for (unsigned iS = 0; iS < s.size(); iS++) {
            if (s[iS]->lhs()) {
               s[iS]->addr(ramAddr);
               ramAddr += s[iS]->bytes();
            }
            else {
               s[iS]->addr(flhAddr);
               flhAddr += s[iS]->bytes();
            }
         }
      }

      void dump(listFile &lst)
      {
         lst.write(0, "\n\n"
                      "Symbol Table Statistics\n"
                      "-----------------------\n"
                      "Symbols:             %4d objects\n\n", _table.size());
         for (int e = symbol::eNone; e < symbol::eLast; e++) {
            lst.write(0, "%s\n", symbol::expNames[e]);
            lst.write(0, "   Name                   Id  Tp  Ar  Size    Addr\n"
                         "   --------------------  ---  --  --  ----  ------\n");
            for (std::map<CString, symbol *>::iterator i = _table.begin(); i != _table.end(); i++) {
               symbol *s = (*i).second;
               if (s->exposed() == e) {
                  lst.write(0, "   %-20s  %3d  %2d  %2d  %4d  0x%04x\n", s->name(), s->id(), s->type(), s->args(), s->bytes(), s->addr());
                  switch (s->ftype()) {
                     case symbol::fNone:
                        break;
                     case symbol::fARR:
                        lst.write(0, "      ARRAY[%3d]\n", s->size());
                        break;
                     case symbol::fCRV:
                        lst.write(0, "      CURVE\n");
                        break;
                     case symbol::fTBL:
                        lst.write(0, "      TABLE\n");
                        break;
                  }
               }
            }
            lst.write(0, "\n");
         }
      }

   private:
      std::map<CString, symbol *> _table;

      void add(CString name, int id, int nArgs) {
         symbol *s = new symbol(name);
         _table[name] = s;
         s->type(symbol::function);
         s->args(nArgs);
      }
};

static symbolTable *symtab = new symbolTable();

//------------------------------------------------------------------------------

class constant {
   public:
      typedef enum {
         ADCv     = 0,
         COUNTERv = 1,
         TIMERv   = 2,
         PWMv     = 3,
         INPUTv   = 4,
         OUTPUTv  = 5
      };

      constant(CString name, int value)
       : _name (name)
       , _value(value)
      {
      }

      CString name () { return _name;  }
      int     value() { return _value; }

   private:
      CString _name;
      int     _value;
};

class constList {
   public:
      constList()
      {
         add("OFF",         0);
         add("ON",          1);
         add("CHANGES",     0);
         add("GOES_TRUE",   1);
         add("GOES_FALSE", -1);
         add("ADC",         0);
         add("COUNTER",     1);
         add("TIMER",       2);
         add("PWM",         3);
         add("INPUT",       4);
         add("OUTPUT",      5);
      }

      void add(const CString name, int value)
      {
         _const[name] = new constant(name, value);
      }

      constant *find(const CString name)
      {
         if (_const.find(name) != _const.end()) {
            return _const[name];
         }
         return NULL;
      }

   private:
      std::map<CString, constant *> _const;
};

static constList _constList;

//------------------------------------------------------------------------------

static char *DecimalNumeric = "0123456789";
static char *BinaryNumeric  = "01";
static char *HexNumeric     = "0123456789abcdefABCDEF";
static char *AlphaNumeric   = "0123456789abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ_";
static char *Alpha          = "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ_";

#define DirChar       '#'
#define AndChar       '&'
#define OrChar        '|'
#define PlusChar      '+'
#define QuoteChar     '"'
#define MinusChar     '-'
#define MulChar       '*'
#define ModChar       '%'
#define DivChar       '/'
#define LParenChar    '('
#define RParenChar    ')'
#define CommaChar     ','
#define EqualChar     '='
#define LtChar        '<'
#define GtChar        '>'
#define DotChar       '.'
#define NotChar       '!'
#define QChar         '?'
#define ColonChar     ':'
#define SemiChar      ';'
#define CaretChar     '^'
#define TildeChar     '~'
#define LBraceChar    '{'
#define RBraceChar    '}'
#define TermChar      '\0'

class lexer {
public:
   typedef enum {
      StrTok,       // "string"
      IntTok,       // integer
      DblTok,       // double
      IdTok,        // Unknown identifier
      symbolTok,    // Identifier that's in the symbol table
      functionTok,  // Function identifier
      constantTok,  // Symbolic constant

      DotTok,       // .
      BAndTok,      // &
      BOrTok,       // |
      BXorTok,      // ^
      BNotTok,      // ~
      LAndTok,      // &&
      LOrTok,       // ||
      LXorTok,      // ^^
      LNotTok,      // !
      CondTok,      // ?
      ColonTok,     // :
      SemiTok,      // ;
      AddTok,       // +
      SubTok,       // -
      MulTok,       // *
      ModTok,       // %
      DivTok,       // /
      LParenTok,    // (
      RParenTok,    // )
      LBraceTok,    // {
      RBraceTok,    // }
      CommaTok,     // ,
      DeclTok,      // :=
      EqualTok,     // =
      LShiftTok,    // <<
      RShiftTok,    // >>
      EqTok,        // ==
      NeTok,        // !=
      LtTok,        // <
      LeTok,        // <=
      GtTok,        // >
      GeTok,        // >=

      CommentTok,   // //

      ArrTok,
      BlockTok,
      CANTok,
      ConstTok,
      CrvTok,
      EndBlockTok,
      EndInterfaceTok,
      InterfaceTok,
      OutTok,
      PortTok,
      TblTok,

      DirTok,       // Directive

      TermTok,      // EOF

      ErrorTok
   } Token;

   private:
      std::map<CString, Token> _keywordList;

      void initKWlist()
      {
         _keywordList["ARR"]           = ArrTok;
         _keywordList["CAN"]           = CANTok;
         _keywordList["CONST"]         = ConstTok;
         _keywordList["CRV"]           = CrvTok;
         _keywordList["OUT"]           = OutTok;
         _keywordList["PORT"]          = PortTok;
         _keywordList["TBL"]           = TblTok;

         _keywordList["BLOCK"]         = BlockTok;
         _keywordList["END_BLOCK"]     = EndBlockTok;
         _keywordList["INTERFACE"]     = InterfaceTok;
         _keywordList["END_INTERFACE"] = EndInterfaceTok;
      }

   public:
      Token       currentTok;
      constant   *currentCns;
      int         CurrentVar;
      symbol     *currentSym;
      CString     currentStr;
      int         currentInt;
      double      currentDbl;

   //---------------------------------------------------------------------------

   private:
#     define CharIsIn(Char,Set) ((Char)?strchr((Set),(Char)):NULL)
#     define GETN_SKIP          (_e.append(getChar()))
#     define PUTN_SKIP          (currentAppend(_e.append(getChar())))
#     define PREV_CHAR(c)       (putBack(c))
#     define NEXT_CHAR          peekAhead()
#     define SKIP_EOL           { while (NEXT_CHAR != '\n') currentStr += GETN_SKIP; }


      char currentAppend(char c)
      {
         currentStr += c;
         return c;
      }

      char putBack(char c)
      {
         assert(_putback == 0);
         _putback = c;
         _e.backup();
         return c;
      }

      char getChar()
      {
         if (_putback) {
            char   c = _putback;
            _putback = 0;
            return c;
         }

         if (_iBuf >= _nBuf) {
            _nBuf = fread(_buffer, sizeof(char), sizeof(_buffer), _input);
            _iBuf = 0;
         }
         if (_iBuf >= _nBuf) return TermChar;

         if (_buffer[_iBuf] == '\n') {
            if (_list) _list->write(_lineNo, 0, "%3d %s\n", _lineNo, LPCTSTR(_line));
            _line.Empty();
            _lineNo++;
            _list->setLine(_lineNo);
            _e.setLine(_lineNo);
         }

         char c = _buffer[_iBuf++];
         if (_blobbing) _blobStr += c;
         if (c != '\n') _line    += c;
         return c;
      }

      char peekAhead()
      {
         if (_putback) {
            return _putback;
         }
         if (_iBuf >= _nBuf) {
            return putBack(getChar());
         }
         return _buffer[_iBuf];
      }

   //---------------------------------------------------------------------------

   void getSequence(char firstChar, const char *validChars, CString &destStr)
   {
      destStr.Empty();
      char  inputChar = firstChar;
      char *itsThere  = CharIsIn(inputChar, validChars);
      while (itsThere) {
         destStr += inputChar;
         inputChar = NEXT_CHAR;
         if ((itsThere = CharIsIn(inputChar, validChars)) != NULL) GETN_SKIP;
      }
   }

   void getUntil(char untilChar, CString &destStr)
   {
      while (NEXT_CHAR != untilChar) {
         destStr += GETN_SKIP;
      }
      destStr += GETN_SKIP; // Past the closing one.
   }

   void getNumber(char firstChar)
   {
      char     inputChar;
      CString  numStr;
      CString  newStr;
      char    *end;
      int      base      = 10;
      char    *Numeric   = DecimalNumeric;

      currentStr = firstChar;

      if (firstChar == '0') { // Allow for binary numbers with a variant on C's hex: 0b10101
         switch (NEXT_CHAR) {
            case 'b': base =  2; PUTN_SKIP; Numeric = BinaryNumeric; break;
            case 'x': base = 16; PUTN_SKIP; Numeric = HexNumeric;    break;
         }
      }

      getSequence(firstChar, Numeric, numStr);
      currentStr += numStr.Mid(1);

      inputChar = NEXT_CHAR;
      if (!CharIsIn(inputChar, ".eE")) {
         //DiagPrint("   Integer value: >%s<\n", numStr);
         currentTok = IntTok;
         currentInt = strtol(LPCTSTR(numStr), &end, base);
         if (*end) _e.Error("Numeric conversion error, value may have overflowed");
      }
      else {
         if (base != 10) _e.Error("Floating point values may only be base 10");

         if (inputChar == '.') {
            numStr += ".";
            PUTN_SKIP;
            inputChar = PUTN_SKIP;

            getSequence(inputChar, DecimalNumeric, newStr);
            if (!newStr.IsEmpty())
               numStr += newStr;
            else
               PREV_CHAR(inputChar);
         }
         if (numStr.GetLength() == 1 && numStr[0] == '.') _e.Error("Expected numeric digits after decimal");

         inputChar = PUTN_SKIP;
         if (CharIsIn(inputChar, "Ee")) {
            numStr += "e";
            inputChar = PUTN_SKIP;

            if (inputChar == '+' || inputChar == '-') {
               if (inputChar == '-') numStr += "-";
               inputChar = PUTN_SKIP;
            }

            getSequence(inputChar, DecimalNumeric, newStr);
            numStr += newStr;
            if (newStr.GetLength() == 0) _e.Error("Expected number for exponent");
            inputChar = PUTN_SKIP;
         }

         PREV_CHAR(inputChar);

         //DiagPrint("   Real value: >%s<\n", numStr);
         currentTok = DblTok;
         currentDbl = strtod(LPCTSTR(numStr), &end);
         if (*end) _e.Error("Numeric conversion error, value may have overflowed");
      }
   }

public:

   lexer(const char *fileName, exprError &ec)
    : _blobbing(false)
    , _e       (ec)
    , _iBuf    (1)
    , _line    ("")
    , _lineNo  (1)
    , _list    (NULL)
    , _nBuf    (0)
    , _putback (0)

   {
      _fileName = fileName;
      _input    = fopen(LPCTSTR(_fileName), "r");

      if (_input == NULL) {
         _e.reset(_fileName);
         _e.Error("Couldn't open file");
      }
      _e.setFile(LPCTSTR(_fileName));
      _e.setLine(_lineNo);

      initKWlist();
   }

  ~lexer()
   {
      fclose(_input);
   }

   void setListFile(listFile *list)
   {
      _list = list;
   }

   //---------------------------------------------------------------------------

   void getToken()
   {
   do {
      while (isspace(NEXT_CHAR)) getChar();
      currentStr.Empty();

      char inputChar = GETN_SKIP;
      _e.reset(inputChar);

      if (CharIsIn(inputChar, Alpha)) {
         getSequence(inputChar, AlphaNumeric, currentStr);
         if ((currentCns = _constList.find(currentStr)) != NULL) {
            currentTok = constantTok;
            currentInt = currentCns->value();
         }
         else if (_keywordList.find(currentStr) != _keywordList.end()) {
            currentTok = _keywordList[currentStr];
         }
         else if ((currentSym = symtab->lookup(currentStr)) != NULL) {
            if (currentSym->type() == symbol::function)
               currentTok = functionTok;
            else
               currentTok = symbolTok;
         }
         else {
            currentTok = IdTok;
         }
      }
      else if (CharIsIn(inputChar, DecimalNumeric)) {
         getNumber(inputChar);
      }
      else {
         currentStr = inputChar;
         switch (inputChar) {
            case QuoteChar : getUntil(QuoteChar, currentStr); currentTok = StrTok; break;
            case DirChar   : currentTok = DirTok;    break;
            case PlusChar  : currentTok = AddTok;    break;
            case MinusChar : currentTok = SubTok;    break;
            case MulChar   : currentTok = MulTok;    break;
            case DotChar   : currentTok = DotTok;    break;
            case ModChar   : currentTok = ModTok;    break;
            case LParenChar: currentTok = LParenTok; break;
            case RParenChar: currentTok = RParenTok; break;
            case LBraceChar: currentTok = LBraceTok; break;
            case RBraceChar: currentTok = RBraceTok; break;
            case CommaChar : currentTok = CommaTok;  break;
            case TermChar  : currentTok = TermTok;   break;
            case TildeChar : currentTok = BNotTok;   break;
            case QChar     : currentTok = CondTok;   break;
            case SemiChar  : currentTok = SemiTok;   break;
            case CaretChar :
               switch (NEXT_CHAR) {
                  case CaretChar : currentTok = LXorTok;   PUTN_SKIP; break;
                  default        : currentTok = BXorTok;              break;
               }
               break;
            case ColonChar :
               switch (NEXT_CHAR) {
                  case EqualChar : currentTok = DeclTok;   PUTN_SKIP; break;
                  default        : currentTok = ColonTok;             break;
               }
               break;
            case DivChar   :
               switch (NEXT_CHAR) {
                  case DivChar   : currentTok = CommentTok; SKIP_EOL; break;
                  default        : currentTok = DivTok;               break;
               }
               break;
            case AndChar   :
               switch (NEXT_CHAR) {
                  case AndChar   : currentTok = LAndTok;   PUTN_SKIP; break;
                  default        : currentTok = BAndTok;              break;
               }
               break;
            case OrChar    :
               switch (NEXT_CHAR) {
                  case OrChar    : currentTok = LOrTok;    PUTN_SKIP; break;
                  default        : currentTok = BOrTok;               break;
               }
               break;
            case NotChar :
               switch (NEXT_CHAR) {
                  case EqualChar : currentTok = NeTok;     PUTN_SKIP; break;
                  default        : currentTok = LNotTok;              break;
               }
               break;
            case EqualChar :
               switch (NEXT_CHAR) {
                  case EqualChar : currentTok = EqTok;     PUTN_SKIP; break;
                  default        : currentTok = EqualTok;             break;
               }
               break;
            case LtChar    :
               switch (NEXT_CHAR) {
                  case LtChar    : currentTok = LShiftTok; PUTN_SKIP; break;
                  case EqualChar : currentTok = LeTok;     PUTN_SKIP; break;
                  default        : currentTok = LtTok;                break;
               }
               break;
            case GtChar    :
               switch (NEXT_CHAR) {
                  case GtChar    : currentTok = RShiftTok; PUTN_SKIP; break;
                  case EqualChar : currentTok = GeTok;     PUTN_SKIP; break;
                  default        : currentTok = GtTok;                break;
               }
               break;

            default        : _e.Error("Invalid input token"); break;
         }
      }
   } while (currentTok == CommentTok);
//_list->write(_lineNo, 1, "LEX %d >%s<\n", currentTok, LPCTSTR(currentStr));
   } 

   void getBlob(Token endToken)
   {
      _blobStr.Empty();
      _blobbing = true;
      do {
         try { getToken(); } catch (exprError) { }
         if (currentTok == TermTok) _e.Error("Unexpected end-of-file inside block");
      } while (currentTok != endToken);
      _blobbing = false;
      currentStr = _blobStr.Mid(0, _blobStr.GetLength()-currentStr.GetLength());
   }

   void getLine()
   {
      _blobStr.Empty();
      _blobbing = true;
      SKIP_EOL;
      _blobbing = false;
      currentStr = _blobStr;
   }

   private:
      bool         _blobbing;
      CString      _blobStr;
      char         _buffer[1024];
      exprError   &_e;
      CString      _fileName;
      int          _iBuf;
      FILE        *_input;
      CString      _line;
      int          _lineNo;
      listFile    *_list;
      int          _nBuf;
      char         _putback;

};

//------------------------------------------------------------------------------

class codeGen {
   public:
      codeGen(const char *fileName, exprError &ec)
       : _err  (ec)
       , _pc   (0)
       , _sp   (0)
       , _maxSP(0)
      {
      }

      typedef enum {
         add,
         sub,
         mul,
         div,
         neg,

         lnot,
         lxor,
         land,
         lor,

         eqr,
         ner,
         ltr,
         ler,
         gtr,
         ger,
         hyst,

         pshi,
         pshd,
         pshv,
         call,
         stov,

         halt,

      } Operator;


      void emit(int n) { 
         Operator op = pshi;
         adjustSP(_op[op].stackEffect);
         _list->write(2, "   # %03d %02d   %s >%d<\n", _pc, _sp, _op[op].name, n);
         _pc += _op[op].pcEffect;
         _code.push_back(instruction(op, n));
      }
      void emit(double n) { 
         Operator op = pshd;
         adjustSP(_op[op].stackEffect);
         _list->write(2, "   # %03d %02d   %s >%f<\n", _pc, _sp, _op[op].name, n);
         _pc += _op[op].pcEffect;
         _code.push_back(instruction(op, n));
      }
      void emit(constant *c) {
         Operator op = pshi;
         adjustSP(_op[op].stackEffect);
         _list->write(2, "   # %03d %02d   %s >%d<  (%s)\n", _pc, _sp, _op[op].name, c->value(), LPCTSTR(c->name()));
         _pc += _op[op].pcEffect;
         _code.push_back(instruction(op, c->value()));
      }
      void emit(Operator op, symbol *s=NULL, int args=0)
      {
         if (op == call) { adjustSP(1); adjustSP(-1); }
         adjustSP(_op[op].stackEffect - args);
         if (s)
            _list->write(2, "   # %03d %02d   %s >%s< %d\n", _pc, _sp, _op[op].name, s->name(), s->id());
         else
            _list->write(2, "   # %03d %02d   %s\n", _pc, _sp, _op[op].name);
         _pc += _op[op].pcEffect;
         if (s)
            _code.push_back(instruction(op, s));
         else
            _code.push_back(instruction(op));
      }

      void resetStack(bool throwError=true)
      {
         if (throwError && _sp != 0) _err.Error("Stack error, too many operands.");
         _sp = 0;
      }

      void setListFile(listFile *list)
      {
         _list = list;
      }

      int codeSize  () { return _pc;    }
      int stackDepth() { return _maxSP; }

      void writeInstructions(const CString &baseName)
      {
         FILE *s19 = fopen(LPCTSTR((baseName+".s19")), "w");
         for (unsigned i = 0; i < _code.size(); i++) {
            instruction &in = _code[i];
            switch(in.t) {
               case instruction::non: fprintf(s19, "%03d %-6s\n",    i+1, _op[in.opCode].name); break;
               case instruction::sym: fprintf(s19, "%03d %-6s %s\n", i+1, _op[in.opCode].name, in.s->name()); break;
               case instruction::nnt: fprintf(s19, "%03d %-6s %d\n", i+1, _op[in.opCode].name, in.n); break;
               case instruction::dbl: fprintf(s19, "%03d %-6s %f\n", i+1, _op[in.opCode].name, in.d); break;
            }
         }
         fclose(s19);
      }

   private:
      listFile      *_list;
      exprError     &_err;
      int            _pc;
      int            _sp;
      int            _maxSP;

      struct instruction {
         typedef enum { non, sym, nnt, dbl } opType;
         Operator  opCode;
         opType    t;
         symbol   *s;
         int       n;
         double    d;
         instruction(Operator o           ) : opCode(o), t(non), s(NULL), n(0), d(0) { }
         instruction(Operator o, symbol *s) : opCode(o), t(sym), s(s),    n(0), d(0) { }
         instruction(Operator o, int     n) : opCode(o), t(nnt), s(NULL), n(n), d(0) { }
         instruction(Operator o, double  d) : opCode(o), t(dbl), s(NULL), n(0), d(d) { }
      };
      std::vector<instruction> _code;

      struct operatorInfo {
         const char *name;
         int         stackEffect; // Change in stack from execution
         int         pcEffect;    // Size of code for this operation
      };
      static operatorInfo  _op[];

      void adjustSP(int n)
      {
         _sp += n;
         if (_sp > _maxSP) _maxSP = _sp;
      }
};

codeGen::operatorInfo codeGen::_op[] = {
   { "ADD",  -1, 1 },
   { "SUB",  -1, 1 },
   { "MUL",  -1, 1 },
   { "DIV",  -1, 1 },
   { "NEG",   0, 1 },

   { "LNOT",  0, 1 },
   { "LXOR", -1, 1 },
   { "LAND", -1, 1 },
   { "LOR",  -1, 1 },

   { "EQ",   -1, 1 },
   { "NE",   -1, 1 },
   { "LT",   -1, 1 },
   { "LE",   -1, 1 },
   { "GT",   -1, 1 },
   { "GE",   -1, 1 },
   { "HYST", -1, 1 },

   { "PSHI",  1, 3 },
   { "PSHD",  1, 3 },
   { "PSHV",  1, 3 },
   { "CALL",  1, 3 },
   { "STOR", -1, 3 },

   { "HALT",  0, 1 },
};

//------------------------------------------------------------------------------

class parser {
   public:
      parser(const char *fileName)
       : _err()
       , _lex(fileName, _err)
       , _lst(fileName)
       , _cdg(fileName, _err)
      {
         _lst.write(99, 0, "File: %s\t%s\n\n", fileName, timeStamp());
         _lex.setListFile(&_lst);
         _cdg.setListFile(&_lst);
      }

      //------------------------------------------------------------------------

      void checkComma(bool scan=false)
      {
         if (_lex.currentTok != lexer::CommaTok) _err.Error("Expected ',' but found");
         if (scan) _lex.getToken();
      }

      void checkLParen(bool scan=false)
      {
         if (_lex.currentTok != lexer::LParenTok) _err.Error("Expected '(', but found");
         if (scan) _lex.getToken();
      }

      void checkRParen(bool scan=false)
      {
         if (_lex.currentTok != lexer::RParenTok) _err.Error("Expected ')', but found");
         if (scan) _lex.getToken();
      }

      int checkInt()
      {
         bool neg = false;
         if (_lex.currentTok == lexer::SubTok ) {
            neg = true;
            _lex.getToken();
         }
         if (_lex.currentTok != lexer::IntTok && _lex.currentTok != lexer::constantTok) _err.Error("Expected integer, but found");
         int n = neg ? -_lex.currentInt : _lex.currentInt;
         _lex.getToken();
         return n;
      }

      double checkDbl()
      {
         bool neg = false;
         if (_lex.currentTok == lexer::SubTok ) {
            neg = true;
            _lex.getToken();
         }
         if (_lex.currentTok != lexer::DblTok && _lex.currentTok != lexer::IntTok) _err.Error("Expected float, but found");
         double n;
         if (_lex.currentTok != lexer::DblTok)
            n = neg ? -_lex.currentDbl : _lex.currentDbl;
         else
            n = neg ? -_lex.currentInt : _lex.currentInt;
         _lex.getToken();
         return n;
      }

      CString checkStr()
      {
         if (_lex.currentTok != lexer::StrTok) _err.Error("Expected string, but found");
         CString s = _lex.currentStr;
         _lex.getToken();
         return s;
      }

      symbol *checkSym()
      {
         if (_lex.currentTok != lexer::symbolTok && _lex.currentTok != lexer::functionTok) _err.Error("Expected symbol, but found");
         symbol *s = _lex.currentSym;
         _lex.getToken();
         return s;
      }

      codeGen::Operator checkOp()
      {
         codeGen::Operator op;
         switch (_lex.currentTok) {
            case lexer::MulTok:  op = codeGen::mul;  break;
            case lexer::DivTok:  op = codeGen::div;  break;
            case lexer::AddTok:  op = codeGen::add;  break;
            case lexer::SubTok:  op = codeGen::sub;  break;
            case lexer::LNotTok: op = codeGen::lnot; break;
            case lexer::LAndTok: op = codeGen::land; break;
            case lexer::LOrTok:  op = codeGen::lor;  break;
            case lexer::LXorTok: op = codeGen::lxor; break;
            case lexer::EqTok:   op = codeGen::eqr;  break;
            case lexer::NeTok:   op = codeGen::ner;  break;
            case lexer::LtTok:   op = codeGen::ltr;  break;
            case lexer::LeTok:   op = codeGen::ler;  break;
            case lexer::GtTok:   op = codeGen::gtr;  break;
            case lexer::GeTok:   op = codeGen::ger;  break;

            default: _err.Error("Expected operator, but found");
         }

         _lex.getToken(); // Past "*" or "/"
         return op;
      }

      //------------------------------------------------------------------------

      void argList(int nArgs)
      {
         checkLParen();
         int args = 0;
         do {
            _lex.getToken(); // Past "(" first, then "," afterward
            expression();
            args++;
         } while (_lex.currentTok == lexer::CommaTok);
         checkRParen(true);
         if (args != nArgs) _err.Error("Expected %d arguments, but found %d", args, nArgs);
      }

      void aFactor()
      {
         bool NOT = false;
         bool NEG = false;
         while (_lex.currentTok == lexer::LNotTok || _lex.currentTok == lexer::SubTok) {
            if (_lex.currentTok == lexer::LNotTok) NOT = !NOT;
            if (_lex.currentTok == lexer::SubTok ) NEG = !NEG;
            _lex.getToken(); // Past unary operator
         }

         switch (_lex.currentTok) {
            case lexer::symbolTok:
               _cdg.emit(codeGen::pshv, checkSym());
               break;

            case lexer::IdTok:
               _err.Error("Unknown identifier");
               break;

            case lexer::constantTok:
               _cdg.emit(_lex.currentCns);
               _lex.getToken();
               break;

            case lexer::IntTok:
               if (NEG) { _lex.currentInt = -_lex.currentInt; NEG = false; }
               _cdg.emit(_lex.currentInt);
               _lex.getToken(); // Past int
               break;

            case lexer::DblTok:
               if (NEG) { _lex.currentDbl = -_lex.currentDbl; NEG = false; }
               _cdg.emit(_lex.currentDbl);
               _lex.getToken(); // Past double
               break;

            case lexer::LParenTok:
               checkLParen(true);
               expression();
               checkRParen(true);
               break;

            case lexer::functionTok:
            {
               symbol *s = _lex.currentSym;
               _lex.getToken(); // Past function name
               argList(s->args());
//             if (s->ftype() == codeGen::fARR, CRV or TBL) {
//                _cdg.emit(codeGen::psha, s);
//                _cdg.emit(codeGen::call, func, s->args()+1);
//             }
//             else
               _cdg.emit(codeGen::call, s, s->args()/*+1*/);
               break;
            }

            default: _err.Error("Unexpected factor");
         }

         if (NEG) _cdg.emit(codeGen::neg);
         if (NOT) _cdg.emit(codeGen::lnot);
      }

      void aTerm()
      {
         aFactor();
         while (_lex.currentTok == lexer::MulTok || _lex.currentTok == lexer::DivTok) {
            codeGen::Operator op = checkOp();
            aFactor();
            _cdg.emit(op);
         }
      }

      void aExpression()
      {
         aTerm();
         while (_lex.currentTok == lexer::AddTok || _lex.currentTok == lexer::SubTok) {
            codeGen::Operator op = checkOp();
            aTerm();
            _cdg.emit(op);
         }
      }

      void relOp()
      {
         aExpression();

         switch (_lex.currentTok) {
            case lexer::EqTok: break;
            case lexer::NeTok: break;
            case lexer::LtTok: break;
            case lexer::LeTok: break;
            case lexer::GtTok: break;
            case lexer::GeTok: break;
            default          : return;
         }
         codeGen::Operator op = checkOp();

         aExpression();
         if (_lex.currentTok == lexer::ColonTok) {
            _lex.getToken(); // Past ":"
            expression();
            _cdg.emit(codeGen::hyst);
         }
         _cdg.emit(op);
      }

      void logFactor()
      {
         relOp();
         while (_lex.currentTok == lexer::LAndTok) {
            codeGen::Operator op = checkOp();
            relOp();
            _cdg.emit(op);
         }
      }

      void logTerm()
      {
         logFactor();
         while (_lex.currentTok == lexer::LOrTok || _lex.currentTok == lexer::LXorTok) {
            codeGen::Operator op = checkOp();
            logFactor();
            _cdg.emit(op);
         }
      }

      void expression()
      {
         logTerm();
      }

      //------------------------------------------------------------------------

      void initialize(symbol *s)
      {
         int n;
         do {
            _lex.getToken(); // Past "{" or ","
            if (_lex.currentTok == lexer::LBraceTok)
               initialize(s);
            else
               n = checkInt();
         } while (_lex.currentTok == lexer::CommaTok);
         _lex.getToken(); // Past "}"
      }

      void declaration(symbol *s)
      {
         do {
            _lex.getToken(); // Past ":=" or ","
            switch (_lex.currentTok) {
               case lexer::CANTok   :
               {
                  s->lhs(true);
                  _lex.getToken(); // "CAN"
                  checkLParen(true);
                  int canID      = checkInt(); checkComma (true);
                  int boardType  = checkInt(); checkComma (true);
                  int dataOffset = checkInt(); checkRParen(true);
                  break;
               }

               case lexer::PortTok  :
               {
                  s->lhs(true);
                  _lex.getToken(); // "PORT"
                  checkLParen(true);
                  int portID   = checkInt(); checkComma(true);
                  int portType = checkInt();

                  switch (portType) {
                     case constant::ADCv:
                        break;

                     case constant::COUNTERv:
                     {
                        checkComma(true); double debouncePeriod = checkDbl();
                        checkComma(true); int    ticksPerCount  = checkInt();
                        break;
                     }

                     case constant::TIMERv:
                     {
                        checkComma(true); double debouncePeriod = checkDbl();
                        break;
                     }

                     case constant::PWMv:
                     {
                        checkComma(true); double frequencyPeriod  = checkDbl();
                        checkComma(true); int    initialDutycycle = checkInt();
                        break;
                     }

                     case constant::INPUTv:
                     {
                        checkComma(true); double debouncePeriod = checkDbl();
                        break;
                     }

                     case constant::OUTPUTv:
                     {
                        checkComma(true); int initialState = checkInt();
                        break;
                     }
                  }
                  checkRParen(true);
                  break;
               }

               case lexer::ArrTok   :
               {
                  _lex.getToken(); // "ARR"
                  checkLParen(true);
                  int arraySize = checkInt();
                  checkRParen(true);
                  try { s->ARR(arraySize); }
                  catch (const char *e) { _err.Error(e); }
                  break;
               }

               case lexer::CrvTok   :
               {
                  _lex.getToken(); // "CRV"
                  checkLParen(true);
                  symbol *xBins = checkSym(); checkRParen(true);
                  try { s->CRV(xBins); }
                  catch (const char *e) { _err.Error(e); }
                  break;
               }

               case lexer::TblTok   :
               {
                  _lex.getToken(); // "TBL"
                  checkLParen(true);
                  symbol *xBins = checkSym(); checkComma(true);
                  symbol *yBins = checkSym(); checkRParen(true);
                  try { s->TBL(xBins, yBins); }
                  catch (const char *e) { _err.Error(e); }
                  break;
               }

               case lexer::ConstTok :
               {
                  _lex.getToken(); // "CONST"
                  checkLParen(true);
                  CString units = checkStr(); checkComma(true);
                  double  scale = checkDbl(); checkComma(true);
                  double  trans = checkDbl(); checkComma(true);
                  double  lo    = checkDbl(); checkComma(true);
                  double  hi    = checkDbl(); checkComma(true);
                  int     dig   = checkInt(); checkRParen(true);
                  try { s->CONST(units, scale, trans, lo, hi, dig); }
                  catch (const char *e) { _err.Error(e); }
                  break;
               }

               case lexer::OutTok   :
               {
                  _lex.getToken(); // "OUT"
                  checkLParen(true);
                  CString units = checkStr(); checkComma(true);
                  double  scale = checkDbl(); checkComma(true);
                  double  trans = checkDbl(); checkComma(true);
                  int     dig   = checkInt(); checkRParen(true);
                  try { s->OUT(units, scale, trans, dig); }
                  catch (const char *e) { _err.Error(e); }
                  break;
               }

               case lexer::LBraceTok:
                  initialize(s);
                  break;

               default       :
                  _err.Error("Unrecognized declaration");
                  break;
            }
         } while (_lex.currentTok == lexer::CommaTok);
      }

      //------------------------------------------------------------------------

      void assignment(symbol *s)
      {
         s->lhs(true); // Make sure it is ram-resident
         _lex.getToken(); // Past =
         expression();
         _cdg.emit(codeGen::stov, s);
         _cdg.resetStack();
         _lst.write(4,"\n");
      }

      //------------------------------------------------------------------------

      void declareOrAssign(bool declared)
      {
         symbol *s = declared ? _lex.currentSym : symtab->add(_lex.currentStr);
         _lex.getToken(); // Past symbol
         switch (_lex.currentTok) {
            case lexer::EqualTok: assignment(s); break;
            case lexer::DeclTok : declaration(s); break;
            default             : _err.Error("Expected ':=' or '=' after identifier");
         }
         if (_lex.currentTok != lexer::SemiTok) _err.Error("Unknown junk at end of expression");
      }

      //------------------------------------------------------------------------

      void directive()
      {
         _lex.getToken(); // Past #
         if (_lex.currentStr == "include") {
            _lex.getToken(); // Past include
            if (_lex.currentTok != lexer::StrTok) {
               _err.Error("Expected file name in include, not %s", LPCTSTR(_lex.currentStr));
            }
         }
         else if (_lex.currentStr == "define") {
            _lex.getToken(); // Past define
            CString id = _lex.currentStr;
            _lex.getLine();
            CString def = _lex.currentStr;
            _lst.write(3, "   DIR >%s< >%s<\n", LPCTSTR(id), LPCTSTR(def));
         }
      }

      //------------------------------------------------------------------------

      void block()
      {
         _lex.getToken(); // Past BLOCK
         parse();         // Past namespace
         if (_lex.currentTok != lexer::EndBlockTok) _err.Error("Expected 'END_BLOCK'");
      }

      //------------------------------------------------------------------------

      void parse()
      {
         do {
            try {
               _lex.getToken();
               switch (_lex.currentTok) {
                  case lexer::functionTok:
                  case lexer::symbolTok  : declareOrAssign(true ); break;
                  case lexer::IdTok      : declareOrAssign(false); break;
                  case lexer::DirTok     : directive      ();      break;

                  case lexer::InterfaceTok:
                     _lst.write(3, "   %03d >%s< blob\n", _lex.currentTok, LPCTSTR(_lex.currentStr));
                     _lex.getBlob(lexer::EndInterfaceTok);
                     _lst.write(3, ">>%s<<\n", LPCTSTR(_lex.currentStr));
                     break;

                  case lexer::BlockTok:
                     block();
                     break;

                  case lexer::EndBlockTok:
                     return;

                  case lexer::TermTok:
                     break;

                  default:
                     _lst.write(3, "   %03d >%s<\n", _lex.currentTok, LPCTSTR(_lex.currentStr));
                     _err.Error("Unknown statement");
                     break;
               }
            }
            catch (exprError err) {
               _lst.write(err.errorLine(), 1, CString("ERROR:")+err.errorString()+"\n\n");
               err.display();
               do { _lex.getToken(); } while (_lex.currentTok != lexer::SemiTok && _lex.currentTok != lexer::TermTok);
               _cdg.resetStack(false);
            }
         } while (_lex.currentTok != lexer::TermTok);

         _cdg.emit(codeGen::halt);
         _lst.flush();

         symtab->computeAddresses();
         _cdg.writeInstructions(_lst.baseName());

         _lst.write(0, "\n\n"
                       "Code Generation Statistics\n"
                       "--------------------------\n"
                       "Code size:           %4d bytes\n"
                       "Maximum stack depth: %4d\n", _cdg.codeSize(), _cdg.stackDepth());

         symtab->dump(_lst);

         _lst.flush();
      }

   //---------------------------------------------------------------------------

   private:
      exprError _err;
      lexer     _lex;
      listFile  _lst;
      codeGen   _cdg;
};

//------------------------------------------------------------------------------

char *fileName = NULL;

void parseArgs(int argc, char *argv[])
{
   fileName = argv[1];
}

//------------------------------------------------------------------------------

int main(int argc, char **argv)
{
   try {
      parseArgs(argc, argv);
      parser p(fileName);
      p.parse();
   }
   catch (exprError err) {
      err.display();
   }
   return 0;
}

//------------------------------------------------------------------------------
