#!/usr/bin/env python
#-------------------------------------------------------------------------------

import sys
import os
import time
import re

intType = type(1)
strType = type("x")
tupType = type((1,))

global _Version
_Version = 'GPIO Configuration Generator 1.0'

#-------------------------------------------------------------------------------

def row(base, N):
   s = ""
   for i in range(N):
      s += '%3d' % ((base+1)*(i+1))
      if i < N-1: s += ', '
   return s

def toInt(S, name):
   try   : return int(S)
   except:
      print 'ERROR: %s value not an integer' % name, S
      return 0

def tryInt(S):
   try   : return int(S)
   except: return S

#-------------------------------------------------------------------------------

class OutputFile:
   def __init__(self, fileName, commentOpener='//', commentCloser=''):
      self.commentOpener = commentOpener
      self.commentCloser = commentCloser
      self.fileName      = fileName
      try:
         self.file = open(self.fileName+'0', 'w')
      except:
         self.file = None
         print "WARNING: Could not open file for output, will not re-generate: %s" % fileName
         sys.exit(0) # Exit if we couldn't open the file...
      self.copyright()

   def copyright(self):
      if self.commentOpener == '<!-- ':
         self.write('<?xml version="1.0" encoding="ISO-8859-1"?>')
      year = time.strftime ("%Y", time.localtime())
      self.comment(self.fileName)
      self.comment(' ', False)
      self.comment('This file was generated by %s'%_Version, False)
      self.comment('Any manual changes will be lost.', False)
      self.comment(' ', False)
      self.comment('Copyright (c) %s, Eric Fahlgren & Lance Gardiner.'%year, False)
      self.comment('All Rights Reserved.', False)
      self.comment(' ', False)
      self.comment()

   def write(self, text="", newline=True):
     if self.file is None: return
     self.file.write(text)
     if newline: self.file.write('\n')

   def comment(self, commentText=None, headline=True):
      l = 80-len(self.commentOpener)-len(self.commentCloser)
      if headline:
         self.write (self.commentOpener + '-' * l + self.commentCloser)
      if commentText:
         l -= 8
         self.write(self.commentOpener + '--  %*s  --' % (-l, commentText) + self.commentCloser)

   def close (self):
      if self.file is None: return
      self.file.close ()

#-------------------------------------------------------------------------------

class funcDef:
   f_or      = 'f_or'
   f_and     = 'f_and'
   f_xor     = 'f_xor'
   f_not     = 'f_not'
   f_add     = 'f_add'
   f_sub     = 'f_sub'
   f_mul     = 'f_mul'
   f_div     = 'f_div'
   f_opers   = (f_or, f_and, f_xor, f_not, f_add, f_sub, f_mul, f_div)

   f_lt      = 'f_lt'      #  Relational operators
   f_le      = 'f_le'
   f_gt      = 'f_gt'
   f_ge      = 'f_ge'
   f_eq      = 'f_eq'
   f_ne      = 'f_ne'
   f_conds   = (f_lt, f_le, f_gt, f_ge, f_eq, f_ne)

   f_convert = 'f_convert' #  1 arg   r <- x*s/d + t
   f_curve   = 'f_curve'   #  1 arg   r <- interp(x, curve)
   f_table   = 'f_table'   #  2 args  r <- interp(x, y, table)

   f_args   = (f_conds, f_convert, f_curve, f_table)

   op = {}
   op['|' ] = f_or
   op['&' ] = f_and
   op['^' ] = f_xor
   op['!' ] = f_not
   op['+' ] = f_add
   op['-' ] = f_sub
   op['*' ] = f_mul
   op['/' ] = f_div
   op['<' ] = f_lt
   op['<='] = f_le
   op['>' ] = f_gt
   op['>='] = f_ge
   op['=='] = f_eq
   op['!='] = f_ne

   def __init__(self, type, count):
      self.type  = type
      self.count = count
      self.H     = 0
      self.N     = 0
      self.M     = 0
      self.S     = 1
      self.D     = 1
      self.T     = 0

   def setShape(self, N, M=0):
      self.N = N
      self.M = M

   def setCvt(self, S, D, T):
      self.S = S
      self.D = D
      self.T = T

   def setHyst(self, H):
      self.H = H

   def shapeStr(self):
      s = "%02d" % self.N
      if self.M: s += "x%02d" % self.M
      return s

   def id(self):
      if self.type in self.f_conds:
         return self.type + "%d" % self.H + "_%03d" % self.count
      elif self.type == self.f_convert:
         return self.type + "%d_%d_%d" % (self.S,self.D,self.T) + "_%03d" % self.count
      else:
         return self.type + self.shapeStr() + "_%03d" % self.count

#-------------------------------------------------------------------------------

class funcUse:
   def __init__(self, fd, dest, src1, src2=None):
      self.fd   = fd
      if type(fd) != strType:
         self.seq  = fd.count-1
      else:
         self.seq  = 0
      self.dest = dest
      self.src1 = src1
      self.src2 = src2

#-------------------------------------------------------------------------------

class symbol:
   _varId = -1
   _cnsId = -1

   variable = 'variable'
   constant = 'constant'

   def __init__(self, name, type='constant', id=None):
      self.name = name
      self.type = type
      self.val  = 0
      if id                : self.id = id
      elif type == symbol.variable: self.id = self.varId()
      else                 : self.id = self.cnsId()

   def varId(self):
      symbol._varId += 1
      return symbol._varId

   def cnsId(self):
      symbol._cnsId += 1
      return symbol._cnsId

   def value(self, val):
      self.val = val

#-------------------------------------------------------------------------------

class funcList:
   def __init__(self):
      self.defList = []
      self.useList = []
      self.symDefs = {}
      self.filter  = None

      self.lineNo  = 0

      self.newVar('ticks')
      self.newVar('seconds')
      for i in range( 8): self.newVar('count%01d' % (i+1))
      for i in range(25): self.newVar('port%02d'  % (i+1))
      for i in range(12): self.newVar('can%02d'   % (i+1)) # ???

      self.split = re.compile(r'^(\w+)=(.+)$')

      self.intS = re.compile(r'^\s*(INTERFACE)\s+([A-Za-z]\w*)\s+"(.*)"$')
      self.intE = re.compile(r'^\s*(END)\s+([A-Za-z]\w*)$')

      self.cns = re.compile(r'^(-?\d+)$')
      self.ali = re.compile(r'^(\w+)$')
      self.neg = re.compile(r'^(!)(\w+)$')
      self.opr = re.compile(r'^(-?\w+)([+=\*/|&])(-?\w+)$')
      self.cnd = re.compile(r'^(\w+)([<=>!]+)(\w+):(\w+)$')
      self.cvt = re.compile(r'^cvt\((\w+),(-?\d+),(-?\d+),(-?\d+)\)$')
      self.crv = re.compile(r'^crv\((\w+),(\d+)\)$')
      self.tbl = re.compile(r'^tbl\((\w+),(\w+),(\d+),(\d+)\)$')

   #----------------------------------------------------------------------------

   def syms(self, typ):
      for key in self.symDefs.keys():
         if self.symDefs[key].type == typ:
            yield(self.symDefs[key])

   def newVar(self, var):
      if self.symDefs.has_key(var):
         print 'ERROR: Variable already exists', var
      else:
         self.symDefs[var] = symbol(var, symbol.variable)
      return self.symDefs[var]

   def oldVar(self, var):
      if not self.symDefs.has_key(var):
         print 'ERROR: Variable doesn\'t exists', var
         self.symDefs[var] = symbol(var, symbol.variable)
      return self.symDefs[var]

   def getVar(self, var):
      if not self.symDefs.has_key(var):
         self.symDefs[var] = symbol(var, symbol.variable)
      return self.symDefs[var]

   def newConst(self, val):
      if not val:
         return val

      if type(val) == strType:
         self.oldVar(val)
         return val

      name = 'const%03d' % (symbol._cnsId+1)
      sym = symbol(name, symbol.constant)
      sym.value(val)
      self.symDefs[name] = sym
      return name

   #----------------------------------------------------------------------------

   def defs(self, typ):
      for fd in self.defList:
         if type(typ) == strType and fd.type == typ: yield(fd)
         if type(typ) == tupType and fd.type in typ: yield(fd)

   def addDef(self, type, count):
      fd = funcDef(type, count)
      self.defList.append(fd)
      return fd

   def addUse(self, fn, dest, src1, src2=None):
      self.getVar(dest)
      src1 = self.newConst(tryInt(src1))
      src2 = self.newConst(tryInt(src2))

      fu = funcUse(fn, dest, src1, src2)
      self.useList.append(fu)
      return fu

   #----------------------------------------------------------------------------

   def addConst(self, dst, val):
      val = toInt(val, 'constant')
      print 'cns %-15s %d' % (dst, val)
      if self.symDefs.has_key(dst):
         print 'ERROR: Symbol already exists', dst
      else:
         sym = symbol(dst, symbol.constant)
         sym.value(val)
         self.symDefs[dst] = sym

   #----------------------------------------------------------------------------

   def addAlias(self, dst, src):
      print 'ali %-15s %-15s' % (dst, src)
      if self.symDefs.has_key(dst):
         print 'ERROR: Symbol already exists', dst
      else:
         self.symDefs[dst] = self.oldVar(src)

   #----------------------------------------------------------------------------

   def addOpr(self, dest, type, src1, src2):
      print 'opr %-15s %-15s %-15s %-15s' % (dest, type, src1, src2)
      self.addUse(type, dest, src1, src2)

   #----------------------------------------------------------------------------

   def getCnd(self, type, H):
      for fn in self.defList:
         if fn.type == type and fn.H == H:
            fn.count += 1
            return fn
      fn = funcDefs.addDef(type, 1)
      fn.setHyst(H)
      return fn

   def addCnd(self, dest, type, src1, src2, H):
      H = toInt(H, 'hysteresis')
      print 'cnd %-15s %-15s %-15s %-15s %d' % (dest, type, src1, src2, H)
      fn = self.getCnd(type, H)
      self.addUse(fn, dest, src1, src2)

   #----------------------------------------------------------------------------

   def getCvt(self, S, D, T):
      for fn in self.defList:
         if fn.type == funcDef.f_convert and fn.S == S and fn.D == D and fn.T == T:
            fn.count += 1
            return fn
      fn = funcDefs.addDef(funcDef.f_convert, 1)
      fn.setCvt(S, D, T)
      return fn

   def addCvt(self, dest, src, S, D, T):
      S = toInt(S, 'scale')
      D = toInt(D, 'divider')
      T = toInt(T, 'translate')
      print 'cvt %-15s %-15s %d %d %d' % (dest, src, S, D, T)
      fn = self.getCvt(S, D, T)
      self.addUse(fn, dest, src)

   #----------------------------------------------------------------------------

   def getCrv(self, nPts):
      for fn in self.defList:
         if fn.type == funcDef.f_curve and fn.N == nPts:
            fn.count += 1
            return fn
      fn = funcDefs.addDef(funcDef.f_curve, 1)
      fn.setShape(nPts)
      return fn

   def addCrv(self, dest, src, nPts):
      nPts = toInt(nPts, 'curve point count')
      print 'crv %-15s %-15s %d' % (dest, src, nPts)
      fn = self.getCrv(nPts)
      self.addUse(fn, dest, src)

   #----------------------------------------------------------------------------

   def getTbl(self, nX, nY):
      for fn in self.defList:
         if fn.type == funcDef.f_table and fn.N == nX and fn.M == nY:
            fn.count += 1
            return fn
      fn = funcDefs.addDef(funcDef.f_table, 1)
      fn.setShape(nX, nY)
      return fn

   def addTbl(self, dest, x, y, nX, nY):
      nX = toInt(nX, 'table row count')
      nY = toInt(nY, 'table column count')
      print 'tbl %-15s %-15s %-15s %d %d' % (dest, x, y, nX, nY)
      fn = self.getTbl(nX, nY)
      self.addUse(fn, dest, x, y)

   #----------------------------------------------------------------------------

   def count(self, types):
      n = 0
      for fn in self.defList:
         if fn.type in types:
            n += fn.count
      return n

   #----------------------------------------------------------------------------

   def parse(self, line):
      self.lineNo += 1
      line = re.sub(';.*', '', line)

      #-------------------------------------------------------------------------
      exp = self.intS.search(line)
      if exp:
         print exp.groups()
         return

      exp = self.intE.search(line)
      if exp:
         print exp.groups()
         return

      #-------------------------------------------------------------------------
      line = re.sub('\s', '', line)
      if line is '': return

      #-------------------------------------------------------------------------
      raw = self.split.search(line)
      if not raw:
         print 'ERROR(1):%03d invalid definition' % self.lineNo, line
         return

      dest = raw.group(1)
      expr = raw.group(2)

      exp = self.cns.search(expr)
      if exp:
         self.addConst(dest, exp.group(1))
         return

      exp = self.ali.search(expr)
      if exp:
         self.addAlias(dest, exp.group(1))
         return

      exp = self.neg.search(expr)
      if exp:
         print exp.groups()
         self.addOpr(dest, funcDef.op[exp.group(1)], exp.group(2), None)
         return

      exp = self.opr.search(expr)
      if exp:
         self.addOpr(dest, funcDef.op[exp.group(2)], exp.group(1), exp.group(3))
         return

      exp = self.cnd.search(expr)
      if exp:
         self.addCnd(dest, funcDef.op[exp.group(2)], exp.group(1), exp.group(3), exp.group(4))
         return

      exp = self.cvt.search(expr)
      if exp:
         self.addCvt(dest, exp.group(1), exp.group(2), exp.group(3), exp.group(4))
         return

      exp = self.crv.search(expr)
      if exp:
         self.addCrv(dest, exp.group(1), exp.group(2))
         return

      exp = self.tbl.search(expr)
      if exp:
         self.addTbl(dest, exp.group(1), exp.group(2), exp.group(3), exp.group(4))
         return

      print 'ERROR: invalid expression', expr

   #----------------------------------------------------------------------------

   def assignIds(self):
      pass

   _use = 0x01
   _def = 0x02
   _cns = 0x04
   _var = 0x08
   _all = 0xFF

   #----------------------------------------------------------------------------

   def dump(self, what=_all):
      if what & funcList._def:
         print
         print 'Definitions'
         for tp in funcDef.f_args:
            for fd in self.defs(tp):
               print '%-10s %3d' % (fd.type, fd.count)

      if what & funcList._use:
         print
         print 'Usage'
         for fu in self.useList:
            if type(fu.fd) == strType:
               print "%-10s     %-15s %-15s %-15s" % (fu.fd,              fu.dest, fu.src1, fu.src2)
            else:
               print "%-10s %3d %-15s %-15s %-15s" % (fu.fd.type, fu.seq, fu.dest, fu.src1, fu.src2)

      if what & funcList._cns:
         print
         print 'Constants'
         for id in range(symbol._cnsId+1):
            for sym in self.syms(symbol.constant):
               if sym.id == id:
                  print '%-15s %s %3d %6d' % (sym.name, sym.type, sym.id, sym.val)

      if what & funcList._var:
         print
         print 'Variables'
         for id in range(symbol._varId+1):
            for sym in self.syms(symbol.variable):
               if sym.id == id:
                  print '%-15s %s %3d' % (sym.name, sym.type, sym.id)

   #----------------------------------------------------------------------------

   def funcDefH(self):
      f = OutputFile('funcDef.h')
      f.write('')
      f.write('#define N_STD f_operators')
      f.write('#define N_CND      %6d' % self.count(funcDef.f_conds))
      f.write('#define N_CVT      %6d' % self.count((funcDef.f_convert,)))
      for fn in self.defs(funcDef.f_curve): f.write('#define N_CRV%-5s %6d' % (fn.shapeStr(), fn.count))
      for fn in self.defs(funcDef.f_table): f.write('#define N_TBL%-5s %6d' % (fn.shapeStr(), fn.count))
      f.write('')

      for fn in self.defs(funcDef.f_curve):
         f.write('typedef struct {')
         f.write('   UINT08  functionType;')
         f.write('   UINT08  nPoints;')
         f.write('   SINT16  xValues[%3d];' % fn.N)
         f.write('   SINT16  yValues[%3d];' % fn.N)
         f.write('} s_Func_crv%s;' % fn.shapeStr())
         f.write('')

      for fn in self.defs(funcDef.f_table):
         f.write('typedef struct {')
         f.write('   UINT08  functionType;')
         f.write('   UINT08  nX;')
         f.write('   UINT08  nY;')
         f.write('   SINT16  xBins [%3d];' % fn.N)
         f.write('   SINT16  yBins [%3d];' % fn.M)
         f.write('   SINT16  values[%3d][%3d];' % (fn.N, fn.M))
         f.write('} s_Func_tbl%s;' % fn.shapeStr())
         f.write('')

      f.comment()
      f.write('')
      f.write('typedef struct {')
      f.write('   UINT16          nDefs;')
      f.write('')
      f.write('   UINT08          operators[N_STD     ];')
      f.write('   s_Func_cnd      cnd      [N_CND     ];')
      f.write('   s_Func_cvt      cvt      [N_CVT     ];')
      for fn in self.defs(funcDef.f_curve):
         f.write('   s_Func_crv%s    crv%s    [N_CRV%s   ];' % (fn.shapeStr(), fn.shapeStr(), fn.shapeStr()))
      for fn in self.defs(funcDef.f_table):
         f.write('   s_Func_tbl%s tbl%s [N_TBL%s];'          % (fn.shapeStr(), fn.shapeStr(), fn.shapeStr()))
      f.write('} s_FuncDef;')
      f.write('')
      f.comment()
      f.close()

   #----------------------------------------------------------------------------

   def funcDefInc(self):
      f = OutputFile('funcDef.inc')
      f.write('')
      f.write('   N_STD + N_CVT + N_CND', False)
      for fn in self.defs(funcDef.f_curve): f.write(' + N_CRV'+fn.shapeStr(), False)
      for fn in self.defs(funcDef.f_table): f.write(' + N_TBL'+fn.shapeStr(), False)
      f.write(',')
      f.write('')
      f.write('   { // Standard functions that have no extra data, must be in exact definition order.')
      f.write('      f_or,  f_and, f_xor,          //       fnId=0-2')
      f.write('      f_add, f_sub, f_mul, f_div    //       fnId=3-7')
      f.write('   },')
      f.write('')
      iFunc = 8

      first = True
      for fn in self.defs(funcDef.f_conds):
         if first:
            f.write('   { // Conditions')
            first = False
         for i in range(fn.count):
            f.write('      { %s, %3d },                //  %3d, fnId=%3d' % (fn.type, fn.H, i, iFunc+i))
         iFunc += fn.count
      if not first:
         f.write('   },')
         f.write('')

      first = True
      for fn in self.defs(funcDef.f_convert):
         if first:
            f.write('   { // Conversions')
            first = False
         for i in range(fn.count):
            f.write('      { f_convert, %3d, %3d, %3d }, //  %3d, fnId=%3d' % (fn.S, fn.D, fn.T, i, iFunc+i))
         iFunc += fn.count
      if not first:
         f.write('   },')
         f.write('')

      for fn in self.defs(funcDef.f_curve):
         f.write('   { // Curve '+fn.shapeStr())
         for i in range(fn.count):
            f.write('      { f_curve,  %2d,               //  %3d, fnId=%3d' % (fn.N, i, iFunc+i))
            f.write('         { %s },' % row(0, fn.N))
            f.write('         { %s }'  % row(0, fn.N))
            f.write('      }, // 0')
         iFunc += fn.count
         f.write('   },')
         f.write('')

      for fn in self.defs(funcDef.f_table):
         f.write('   { // Table '+fn.shapeStr())
         for i in range(fn.count):
            f.write('      { f_table,  %2d,  %2d,          //  %3d, fnId=%3d' % (fn.N, fn.M, i, iFunc+i))
            f.write('         { %s },' % row(0, fn.N))
            f.write('         { %s },' % row(0, fn.M))
            f.write('         {')
            for n in range(fn.N):
               f.write('            { %s },' % row(n, fn.M))
            f.write('         }')
            f.write('      },')
         iFunc += fn.count
         f.write('   },')
         f.write('')

      f.comment()
      f.close()

   #----------------------------------------------------------------------------

   def funcDefUsrCon(self):
      f = OutputFile('funcDefUsr.con')
      f.comment()
      f.close()

   #----------------------------------------------------------------------------

   def funcDefUsrVar(self):
      f= OutputFile('funcDefUsr.var')
      # Maps the outputs from each function usage, not the definitions.
      """
      N_STD + N_CVT + N_CND + N_CRV + N_TBL
      SINT16 conversion[N_CVT  ]; //  "Conversions"  f(x) = x * s / d + t 
      SINT16 condition [N_CND  ]; //  Conditions
      SINT16 curve8    [N_CRV08]; //   8-point curves
      SINT16 curve16   [N_CRV16]; //  16-point curves
      SINT16 table8x8  [N_TBL08]; //   8x 8 table
      SINT16 table12x12[N_TBL12]; //  12x12 table
      """
      f.comment()
      f.close()

   #----------------------------------------------------------------------------

   def funcDefIniCon(self):
      f= OutputFile('funcDefIni.con', ';')
      f.close()

   #----------------------------------------------------------------------------

   def funcDefIniVar(self):
      f= OutputFile('funcDefIni.var', ';')
      f.close()

   #----------------------------------------------------------------------------

   def funcDefMsq(self):
      f= OutputFile('funcDef.msq', '<!-- ', ' -->')
      f.close()

#-------------------------------------------------------------------------------

if __name__ == '__main__':

   funcDefs = funcList()

   # Predefined are seconds, ticks, count1-8, port01-25 + can01-nn?
   funcDefs.parse('; Global values')
   funcDefs.parse('rpm           = can04')
   funcDefs.parse('map           = cvt(port02, 100, 104,  3)')
   funcDefs.parse('tps           = cvt(port03, 100, 893, -4)')
   funcDefs.parse('clt           = crv(port04, 20)')
   funcDefs.parse('wue           = crv(clt,    16)')

   funcDefs.parse('; Nitrous example.')
   funcDefs.parse('INTERFACE nc "Nitrous Control"')
   funcDefs.parse('   nitrousEnable = port12 ; Assign to port12')
   funcDefs.parse('   armSwitch     = port18')
   funcDefs.parse('   rpmLimit      = 3500')
   funcDefs.parse('END nc')


   funcDefs.parse('ten           = 19 + -9')

   funcDefs.parse('ve1           = tbl(rpm, map, 8, 8)')
   funcDefs.parse('ve2           = tbl(rpm, map, 8, 8)')

   funcDefs.parse('nitrousCond1  = rpm >= rpmLimit:100')
   funcDefs.parse('nitrousCond2  = tps >   75:10')
   funcDefs.parse('nitrousCond3  = clt >  160: 5')
   funcDefs.parse('nitrousCond4  = nitrousCond1 & nitrousCond2')
   funcDefs.parse('nitrousCond   = nitrousCond3 & nitrousCond4')
   funcDefs.parse('nitrousEnable = nitrousCond  & armSwitch')

   funcDefs.assignIds()

   funcDefs.dump(funcList._cns)# | funcList._var)

   if 0:
      f = funcDefs.addDef(funcDef.f_lt,      2) ; f.setHyst(2)
      f = funcDefs.addDef(funcDef.f_gt,      1) ; f.setHyst(0)
      f = funcDefs.addDef(funcDef.f_gt,      1) ; f.setHyst(9)
      f = funcDefs.addDef(funcDef.f_convert, 2) ; f.setCvt(2, 1, 3)
      f = funcDefs.addDef(funcDef.f_curve,   1) ; f.setShape( 8)
      f = funcDefs.addDef(funcDef.f_curve,   1) ; f.setShape(16)
      f = funcDefs.addDef(funcDef.f_table,   2) ; f.setShape( 8,  8)
      f = funcDefs.addDef(funcDef.f_table,   1) ; f.setShape( 6, 12)

   #----------------------------------------------------------------------------
   #--  Write the definition files.  -------------------------------------------
   funcDefs.funcDefH     ()
   funcDefs.funcDefInc   ()
   funcDefs.funcDefUsrCon()
   funcDefs.funcDefUsrVar()
   funcDefs.funcDefIniCon()
   funcDefs.funcDefIniVar()
   funcDefs.funcDefMsq   ()
