//------------------------------------------------------------------------------
//--  Function Evaluator - GPIO Support                                       --
//--                                                                          --
//--  Copyright (c) 2006 by Eric Fahlgren & Lance Gardiner                    --
//--  All Rights Reserved.                                                    --
//--                                                                          --
//--  This program is free software; you can redistribute it and/or           --
//--  modify it under the terms of the GNU General Public License             --
//--  as published by the Free Software Foundation; either version            --
//--  2 of the License, or (at your option) any later version.                --
//--  See http://www.gnu.org/licenses/gpl.txt                                 --
//------------------------------------------------------------------------------

#include "hcs12def.h"
#include "ms2def.h"
#include "function.h"

#include <stdlib.h>

#ifndef GCC_BUILD
#  pragma CODE_SEG DEFAULT
#endif

//------------------------------------------------------------------------------
//--  Function definitions flash segment  --------------------------------------
#ifndef GCC_BUILD
#  pragma CONST_SEG FNC_ROM
#endif

const s_FuncDef funcDef = {
#  include "funcDef.inc" // Include the auto-generated initializations.
};

#define FUNC_DEF(i) ((FunctionDefinition)(((UINT08 *)funcDef.operators)+i))

#ifndef GCC_BUILD
#  pragma CONST_SEG DEFAULT
#endif
//------------------------------------------------------------------------------

static Function  userFunction = NULL;
static UINT08   nFunctions    = 0;

//------------------------------------------------------------------------------

static SINT16 interp(SINT16 x, SINT16 x0, SINT16 y0, SINT16 x1, SINT16 y1)
{
   return y0 + (x-x0) * (y1-y0) / (x1-x0);
}

//------------------------------------------------------------------------------

#define fTYPE      u.c.functionType
#define NPOINTS    u.c.nPoints
#define CPOINTS    u.c.data
#define NX         u.t.nX
#define NY         u.t.nY
#define VALUES     u.t.data
#define SCALE      u.s.scale
#define TRANS      u.s.trans
#define DIVID      u.s.divid
#define HYSTERESIS u.n.hysteresis

static FunctionDefinition DEF(UINT08 iDef)
{
   UINT16 iAddr = 0;

   if (iDef < funcDef.nDefs) {
      UINT08 i;
      for (i = 0; i < iDef; i++) {
         FunctionDefinition f = FUNC_DEF(iAddr);
         iAddr += sizeof(f->fTYPE);
         switch (f->fTYPE) {
            case f_curve:
               iAddr += sizeof(f->NPOINTS) + f->NPOINTS*sizeof(SINT16);
               break;

            case f_table:
               iAddr += sizeof(f->NX) + sizeof(f->NY) + (f->NX*f->NY+f->NX+f->NY)*sizeof(SINT16);
               break;

            case f_convert:
               iAddr += 3*sizeof(SINT16);
               break;

            case f_lt:
            case f_le:
            case f_gt:
            case f_ge:
            case f_eq:
            case f_ne:
               iAddr += sizeof(SINT16);
               break;
         }
      }
   }

   return FUNC_DEF(iAddr);
}

//------------------------------------------------------------------------------

SINT16 func_Interpolate2D(FunctionDefinition func, SINT16 x)
{
   UINT08 ip;

   UINT08 n = func->NPOINTS;

#define IX(i) (i)
#define IY(i) (n+(i))
   if (x <= func->CPOINTS[IX(0)  ]) return func->CPOINTS[IY(0)  ];
   if (x >= func->CPOINTS[IX(n-1)]) return func->CPOINTS[IY(n-1)];

   for (ip = 0; ip < n; ip++) {
      if (x > func->CPOINTS[ip]) break;
   }

   return interp(x, func->CPOINTS[IX(ip)], func->CPOINTS[IX(ip+1)], func->CPOINTS[IY(ip)], func->CPOINTS[IY(ip+1)]); 
#undef IX
#undef IY
}

//------------------------------------------------------------------------------

static SINT16 XBIN(FunctionDefinition func, UINT08 i) {
   return func->VALUES[i];
}
static SINT16 YBIN(FunctionDefinition func, UINT08 i) {
   return func->VALUES[func->NX+i];
}
static SINT16 ZVAL(FunctionDefinition func, UINT08 ix, UINT08 iy) {
   return func->VALUES[func->NX+func->NY+ix*func->NY+iy];
}

SINT16 func_Interpolate3D(FunctionDefinition func, SINT16 x, SINT16 y)
{
   UINT08 iX,  iY;
   SINT16 X0,  X1;
   SINT16 Y0,  Y1;
   SINT16 Z0,  Z1;
   SINT16 Z00, Z10, Z01, Z11;

   UINT08 lastX = func->NX-1;
   UINT08 lastY = func->NY-1;

   if (x < XBIN(func, 0    )) x = XBIN(func, 0);
   if (x > XBIN(func, lastX)) x = XBIN(func, lastX);
   if (y < YBIN(func, 0    )) y = YBIN(func, 0);
   if (y > YBIN(func, lastY)) y = YBIN(func, lastY);

   for (iX = 0; iX < lastX; iX++) {
      if (x <= XBIN(func, iX+1)) {
         X0 = XBIN(func, iX);
         X1 = XBIN(func, iX+1);
         break;
      }
   }
   for (iY = 0; iY < lastY ; iY++) {
      if (y <= YBIN(func, iY+1)) {
         Y0 = YBIN(func, iY);
         Y1 = YBIN(func, iY+1);
         break;
      }
   }

   Z00 = ZVAL(func, iX+0, iY+0);
   Z10 = ZVAL(func, iX+1, iY+0);
   Z01 = ZVAL(func, iX+0, iY+1);
   Z11 = ZVAL(func, iX+1, iY+1);

   Z0 = interp(x, X0, Z00, X1, Z10);
   Z1 = interp(x, X0, Z01, X1, Z11);
   return interp(y, Y0, Z0, Y1, Z1);
}

//------------------------------------------------------------------------------

SINT16 func_Convert(FunctionDefinition func, SINT16 x)
{
   return x * func->SCALE / func->DIVID + func->TRANS;
}

//------------------------------------------------------------------------------

SINT16 func_GT(FunctionDefinition func, SINT16 x, SINT16 y, SINT16 state)
{
   SINT16 d = x-y;
   if (d > 0)
      state = 1;
   else if (state && d <= -func->HYSTERESIS)
      state = 0;
   return state;
}

SINT16 func_GE(FunctionDefinition func, SINT16 x, SINT16 y, SINT16 state) { return func_GT(func, x,   y-1, state); }
SINT16 func_LT(FunctionDefinition func, SINT16 x, SINT16 y, SINT16 state) { return func_GT(func, y,   x,   state); }
SINT16 func_LE(FunctionDefinition func, SINT16 x, SINT16 y, SINT16 state) { return func_GT(func, y+1, x,   state); }

SINT16 func_EQ(FunctionDefinition func, SINT16 x, SINT16 y, SINT16 state)
{
   SINT16 d = abs(x-y);
   if (d == 0)
      state = 1;
   else if (state && d > func->HYSTERESIS)
      state = 0;
   return state;
}

SINT16 func_NE(FunctionDefinition func, SINT16 x, SINT16 y, SINT16 state) { return !func_EQ(func, x, y, !state); }

//------------------------------------------------------------------------------

void func_EvalOne(UINT08 iFunc)
{
   Function f = userFunction+iFunc;
   if (f->output != NULL) {
      switch (f->func->fTYPE) {
         case f_convert: *f->output = func_Convert      (f->func, *f->input1);             break;
         case f_curve  : *f->output = func_Interpolate2D(f->func, *f->input1);             break;
         case f_table  : *f->output = func_Interpolate3D(f->func, *f->input1, *f->input2); break;

         case f_lt     : *f->output = func_LT(f->func, *f->input1, *f->input2, *f->output); break;
         case f_le     : *f->output = func_LE(f->func, *f->input1, *f->input2, *f->output); break;
         case f_gt     : *f->output = func_GT(f->func, *f->input1, *f->input2, *f->output); break;
         case f_ge     : *f->output = func_GE(f->func, *f->input1, *f->input2, *f->output); break;
         case f_eq     : *f->output = func_EQ(f->func, *f->input1, *f->input2, *f->output); break;
         case f_ne     : *f->output = func_NE(f->func, *f->input1, *f->input2, *f->output); break;

         case f_or     : *f->output = *f->input1 != 0 || *f->input2 != 0; break;
         case f_and    : *f->output = *f->input1 != 0 && *f->input2 != 0; break;
         case f_xor    : *f->output = *f->input1 != 0 ^  *f->input2 != 0; break;

         case f_add    : *f->output = *f->input1      +  *f->input2;      break;
         case f_sub    : *f->output = *f->input1      -  *f->input2;      break;
         case f_mul    : *f->output = *f->input1      *  *f->input2;      break;
         case f_div    : *f->output = *f->input1      /  *f->input2;      break;
      }
   }
}

//------------------------------------------------------------------------------

void func_Evaluate(void)
{
   UINT08 iF;
   for (iF = 0; iF < nFunctions; iF++) {
      func_EvalOne(iF);
   }
}

//------------------------------------------------------------------------------

void func_Add(UINT08 iDef, SINT16 *out, SINT16 *in1, SINT16 *in2)
{
   static UINT08 iFunc = 0;
   if (iFunc < nFunctions) {
      Function f = userFunction+iFunc;
      f->func    = DEF(iDef);
      f->input1  = in1;
      f->input2  = in2; // Null for 2d curves.
      f->output  = out;
      iFunc++;
   }
}

//------------------------------------------------------------------------------

void func_Init(UINT08 nF)
{
   nFunctions   = nF;
   userFunction = (Function)calloc(nFunctions, sizeof(Function));
}

//------------------------------------------------------------------------------
