//------------------------------------------------------------------------------
//--  Input Capture - GPIO Support                                            --
//--                                                                          --
//--  Copyright (c) 2006 by Eric Fahlgren & Lance Gardiner                    --
//--  All Rights Reserved.                                                    --
//--                                                                          --
//--  This program is free software; you can redistribute it and/or           --
//--  modify it under the terms of the GNU General Public License             --
//--  as published by the Free Software Foundation; either version            --
//--  2 of the License, or (at your option) any later version.                --
//--  See http://www.gnu.org/licenses/gpl.txt                                 --
//------------------------------------------------------------------------------

#include "hcs12def.h"
#include "ms2def.h"
#include "inputCapture.h"

#ifndef GCC_BUILD
#  pragma CODE_SEG DEFAULT
#endif

//------------------------------------------------------------------------------

extern UINT32 tix;

static s_inputCapture ic[N_TIMERS];

//------------------------------------------------------------------------------

void inputCapture_init(void)
{
   UINT08 i;
   for (i = 0; i < N_TIMERS; i++) {
      ic[i].isEnabled     = FALSE;
      ic[i].mask          = 0x01 << i;
      ic[i].timerOutput   = NULL;
      ic[i].countOutput   = NULL;

      ic[i].time          = 0;
      ic[i].debounce      = TRUE;
      ic[i].reenableTime  = 0;
      ic[i].events        = 0;
      ic[i].eventInterval = 1;
   }
}

//------------------------------------------------------------------------------

void inputCapture_enable(UINT08 iPort, UINT16 p, SINT16 *t, UINT16 *c)
{
   ic[iPort].isEnabled     = TRUE;
   ic[iPort].timerOutput   = t;
   ic[iPort].countOutput   = c;
   ic[iPort].eventInterval = p;

   DDRT   &= ~ic[iPort].mask;
   TCTL34 |= (UINT16)ic[iPort].mask << iPort;
   TFLG1  |=  ic[iPort].mask;
   TIE    |=  ic[iPort].mask;
   TIOS   &= ~ic[iPort].mask;
}

//------------------------------------------------------------------------------
//--  The re-enable logic should handle three cases:                          --
//--                                                                          --
//--     1) Tooth wheel with regular spacing, such as ABS wheel sensors.      --
//--     2) Tooth wheel with missing teeth, e.g., crank triggers.             --
//--     3) Arbitrary intervals, such as toggle switches.                     --
//--                                                                          --
//--  Case 1 is handled will by using a proportional debounce period,         --
//--  some percentage of the previous time interval, but this sometimes       --
//--  causes missed teeth for case 2.  Case three is best solved using a      --
//--  fixed interval debounce period, typically a lot longer for mechanical   --
//--  switches than for electronic (Hall or VR) ones.                         --
//------------------------------------------------------------------------------

void inputCapture_capture(UINT08 iPort)
{
   *ic[iPort].timerOutput = ic[iPort].time;
   if (ic[iPort].debounce) {
      UINT16 debouncePeriod  = ic[iPort].time >> 2; // ??? FIX THIS, allow other than 25% debounce period
      ic[iPort].reenableTime = tix + debouncePeriod;
      TIE &= ~ic[iPort].mask;
   }
   ic[iPort].time = 0;

   ic[iPort].events++;
   if (ic[iPort].events > ic[iPort].eventInterval) {
      (*ic[iPort].countOutput)++;
      ic[iPort].events = 0;
   }

   TFLG1 |= ic[iPort].mask;
}

//------------------------------------------------------------------------------
//--  Must be called in clock ISR to re-enable interrupts.  --------------------

void inputCapture_debounce(void)
{
   UINT08 i;
   for (i = 0; i < N_TIMERS; i++) {
      if (ic[i].isEnabled) {
         ic[i].time++;
         if (ic[i].debounce && tix > ic[i].reenableTime) {
            TIE |=  ic[i].mask;
         }
      }
   }
}

//------------------------------------------------------------------------------
